/*
   Copyright (C) 2003-2007  Mathias Broxvall

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <SDL.h>
#include "SDL/SDL_opengl.h"
#include <sys/time.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <getopt.h>
#include "zlib.h"
#include <pthread.h>

#include <X11/X.h>
#include <X11/Xatom.h>
#include <X11/Xlib.h>
#include "vroot.h"

#ifdef WITH_ROOTWINDOW
#include "GL/glu.h"
#include "GL/glx.h"
#endif

/*------------------- CONSTANTS ---------------*/
#define MAX_POINTS 100000
#define MAX_M_WIDTH 1024
#define MAX_M_HEIGHT 1024
#define N_LINKS 5

/*------------------- TYPES -----------------*/
typedef struct Point {
  int index;
  float pos[2];         /* Position of point */
  float dir[2];         /* Speed of point */
  float angular;        /* Angular speed of point */
  float heat;           /* Heat of point */
  short burn_left,is_burning; 
  short links[N_LINKS];
  struct Point *next;   /* List when attached to map */
} Point;

/* A more compact representation of a point suitable for storing to
   disk */
typedef struct DiskPoint {
  float pos[2],heat;
  char burn_left,is_burning;
} DiskPoint;

typedef struct ME {
  Point *points;
  int gravity;
  double heat;
} ME;

/* Store computation work orders for slave threads */
typedef struct WorkOrder {
  int x1,y1,x2,y2;
  struct WorkOrder *next;
} WorkOrder;

/*----------------- GLOBALS -----------------*/
extern FILE *yyin;

SDL_Surface *screen;
Point points[MAX_POINTS];
DiskPoint diskpoints[MAX_POINTS];
ME map[MAX_M_WIDTH][MAX_M_HEIGHT];
int map2[MAX_M_WIDTH/4][MAX_M_HEIGHT/4];
int map3[MAX_M_WIDTH/16][MAX_M_HEIGHT/16];
int silent=1,burn_steps=3;
int npoints=400;
int sWidth=800,sHeight=600;
int is_running;

struct WorkOrder *workQueue=NULL;
struct WorkOrder *freeOrders=NULL;
pthread_mutex_t orderMutex=PTHREAD_MUTEX_INITIALIZER;
int maxOrders=0;

#define MAX_SLAVES 256
pthread_t slaveThreads[MAX_SLAVES];
int numSlaves=4;

/*----------------- Simulation Constants -------------------*/
double gravity=4e-7, repell=-50e-9, friction=3000e-8;
double dist_off_g=20e-3, dist_off_r=10e-6, dist_off_f=5e-3;
double radiation=0.05, transfer=4e-8, friction_heat=100e-2;
double heat_off=0.0,fusion_radiation=10e4,fusion_chance=50e-8,fusion_heat=50.0;
int m_width=128, m_height=128,m_complicated=0;
double m_res=20.0;
double max_link_distance=100e-3,    /* Break link of distance if greater than this */
  link_spring_distance=10e-3,       /* Minimum distance needed for link to act as spring */
  link_spring_constant=1e1,        /* Force applied by link */
  link_temp_prop=0.030,            /* Amount of temperature propgated by link */
  link_creation_distance=10e-3,    /* Links can be created if particles are this close */ 
  link_creation_chance=0.2,        /* This is the chance to create link */
  link_fric=0.1,                   /* Friction caused by links */  
  angular_epsilon=0.1;              /* Angular momentum propagated by links */
double fusion_dist=20e-4;
double cometRadius=0.1;

/* Usefull globals */
double simulationRadius;

struct temperature {
  int n;
  float rgb[5][4];
};

#define NUM_COLORS 18
struct temperature colors[NUM_COLORS] = {
  {1,{{0.3,0.3,0.3,0.5},{0.0,0.0,0.0,0.5},{0.0,0.0,0.0,0.5},{0.0,0.0,0.0,0.5},{0.0,0.0,0.0,0.5}}},
  {1,{{0.3,0.3,0.3,0.5},{0.0,0.0,0.0,0.5},{0.0,0.0,0.0,0.5},{0.0,0.0,0.0,0.5},{0.0,0.0,0.0,0.5}}},
  {1,{{0.3,0.3,0.4,0.5},{0.0,0.0,0.0,0.5},{0.0,0.0,0.0,0.5},{0.0,0.0,0.0,0.5},{0.0,0.0,0.0,0.5}}},
  {1,{{0.1,0.1,0.5,0.5},{0.0,0.0,0.0,0.5},{0.0,0.0,0.0,0.5},{0.0,0.0,0.0,0.5},{0.0,0.0,0.0,0.5}}},
  {2,{{0.3,0.0,0.1,0.5},{0.1,0.0,0.1,0.5},{0.0,0.0,0.0,0.5},{0.0,0.0,0.0,0.5},{0.0,0.0,0.0,0.5}}},
  {2,{{0.5,0.0,0.0,0.5},{0.3,0.0,0.0,0.5},{0.0,0.0,0.0,0.5},{0.0,0.0,0.0,0.5},{0.0,0.0,0.0,0.5}}},
  {2,{{0.6,0.2,0.0,0.5},{0.3,0.0,0.0,0.5},{0.0,0.0,0.0,0.5},{0.0,0.0,0.0,0.5},{0.0,0.0,0.0,0.5}}},
  {3,{{0.7,0.4,0.0,0.5},{0.4,0.1,0.1,0.5},{0.0,0.0,0.1,0.5},{0.0,0.0,0.0,0.5},{0.0,0.0,0.0,0.5}}},
  {3,{{0.8,0.6,0.0,0.5},{0.5,0.3,0.1,0.5},{0.0,0.0,0.1,0.5},{0.0,0.0,0.0,0.5},{0.0,0.0,0.0,0.5}}},
  {3,{{0.9,0.7,0.1,0.5},{0.5,0.5,0.1,0.5},{0.2,0.1,0.0,0.5},{0.0,0.0,0.0,0.5},{0.0,0.0,0.0,0.5}}},
  {3,{{1.0,0.8,0.3,0.5},{0.7,0.6,0.1,0.5},{0.2,0.1,0.0,0.5},{0.0,0.0,0.0,0.5},{0.0,0.0,0.0,0.5}}},
  {4,{{1.0,0.9,0.5,0.5},{0.8,0.7,0.2,0.5},{0.2,0.1,0.0,0.5},{0.1,0.0,0.0,0.5},{0.0,0.0,0.0,0.5}}},
  {4,{{1.2,1.0,0.8,0.5},{0.8,0.7,0.4,0.5},{0.3,0.2,0.0,0.5},{0.2,0.1,0.0,0.5},{0.0,0.0,0.0,0.5}}},
  {4,{{1.5,1.2,0.8,0.5},{1.0,0.9,0.5,0.5},{0.4,0.3,0.1,0.5},{0.3,0.2,0.0,0.5},{0.0,0.0,0.0,0.5}}},
  {4,{{2.0,1.5,1.0,0.5},{1.0,1.0,0.8,0.5},{0.5,0.4,0.2,0.5},{0.4,0.3,0.1,0.5},{0.0,0.0,0.0,0.5}}},
  {5,{{3.0,2.0,1.5,0.5},{1.2,1.2,1.0,0.5},{0.8,0.6,0.3,0.5},{0.5,0.4,0.2,0.5},{0.1,0.0,0.0,0.5}}},
  {5,{{4.0,3.0,2.0,0.5},{1.5,1.5,1.2,0.5},{1.0,0.8,0.4,0.5},{0.6,0.5,0.3,0.5},{0.2,0.1,0.0,0.5}}},
  {5,{{5.0,4.0,2.5,0.5},{1.8,1.8,1.4,0.5},{1.2,1.0,0.5,0.5},{0.7,0.6,0.4,0.5},{0.4,0.2,0.1,0.5}}}
};
#define MAX_BURN_STEPS 7
GLfloat burn_colors[MAX_BURN_STEPS][4]={{1.0,1.0,1.0,1.0}, {1.0,1.0,0.2,1.0}, {1.0,0.2,0.2,1.0}, {1.0,0.2,1.0,1.0}, {0.2,0.2,1.5,1.0}, {0.2,1.0,1.0,1.0}, {0.2,1.0,0.2,1.0}};
  


/*----------------- PROTOTYPES --------------------*/

/* "objects" */
void sphere(int from,int to,double cx,double cy,double rad);
void cube(int from,int to,double cx,double cy,double rad);

/* transformations on them */
void spin(int from,int to,double cx,double cy,double rad,double speed);
void linear(int from,int to,double dx,double dy);
void drift(int from,int to,double speed);
void zero(int from,int to);

/* Special effects */
void createComet();
void createPlanet();

void interact(Point *p1,Point *p2);
void queue_interaction(int,int,int,int);
void *slave_thread_fn(void *);
int log2i(int);
int yyparse();
double frandom();

void printUsage(FILE *stream) {
  fprintf(stream, "Usage: xgrav [-w -f] [-r, -p <file>] -s <settingsFile> \n");
  fprintf(stream, 
	  "    -w      screen width (default 800)\n"
	  "    -h      screen height (default 600)\n"
	  "    -f      Run in fullscreen \n"
	  "    -r      Record simulation\n"
	  "    -p      Play (and continue) simulation\n"
	  "    -s      Settings file\n"
	  "    -l      Draw particle links\n"
	  "    -e      Skip to end during playback\n"
	  "    -t      Number of slave threads to use\n"
	  "    --root  Use root window for display\n"
	  "    --fps   Print number of frame per second\n");
  exit(0);
}


Window window;
Display *display;
int x11screen;
int useRootWindow,screenWidth,screenHeight;


/*                                              */
/*------------------ MAIN ----------------------*/
/*                                              */

int main(int argc,char **argv) {
  int x,y,i,j;
  int x2,y2,c;
  double dx,dy,d,v;
  SDL_Event event;
  SDL_Rect r;
  Point *p1,*p2;
  int frames=0;
  struct timeval tv;
  double t0,t1;
  double zoom=1.0;
  double offset_mx,offset_my;
  int mx2,mx3,my2,my3;
  int mx2_2,mx3_2,my2_2,my3_2;
  double cam[2]={0.0,0.0}; // Camera focus
  int printFPS=0;

  /* options */
  int nextOption,isWindowed=1,drawLinks=0;
  char *filename;
  char *settingsFile=NULL;
  const char *const shortOptions="?w:h:r:p:fs:let:";
  const struct option longOptions[] = {
	{ "help",     0, NULL, '?' },
	{ "width",     1, NULL, 'w' },
	{ "height", 1, NULL, 'h' },
	{ "fullscreen", 0, NULL, 'f' },
	{ "record",   1, NULL, 'r' },
	{ "play",     1, NULL, 'p' },
	{ "root",     0, NULL, 'R' },
	{ "settings",     1, NULL, 's' },
	{ "links",     1, NULL, 'l' },
	{ "window-id",     1, NULL, 'W' },
	{ "end",     0, NULL, 'e' },
	{ "fps",     0, NULL, '!' },
	{ "threads",     1, NULL, 't' },
	{ NULL,       0, NULL, 0 }
  };

  /* file used for recording simulation */
  int record_to_disk=0,replay_from_disk=0,skip_to_end=0;
  FILE *file=NULL;
  useRootWindow=0;
  window=0;

  printf("Welcome to XGrav v%s\n",VERSION);
  printf("Released under the GNU General Public License (GPL). ");
  printf("By Mathias Broxvall. \n");

  do {
	nextOption=getopt_long(argc,argv,shortOptions,longOptions,NULL);
	switch(nextOption) {
	case '?': printUsage (stdout); break;	  
	case 'w': sWidth=atoi(optarg); break;
	case 'h': sHeight=atoi(optarg); break;
	case 'f': isWindowed=0; break;
	case 'r': record_to_disk=1; file=fopen(optarg,"wb"); filename=optarg; break;
	case 'p': replay_from_disk=1; file=fopen(optarg,"rb"); filename=optarg; break;
	case 'R': useRootWindow=1; break;
	case 's': settingsFile=optarg; break;
	case 'W': window=strtol(optarg,0, 0); break;
	case 'l': drawLinks=1; break;
	case 'e': skip_to_end=1; break;
	case '!': printFPS=1; break;
	case 't': numSlaves=atoi(optarg); break;
	}
  }  while(nextOption != -1);
  if(!settingsFile) printUsage(stdout);

  r.x=0; r.y=0; r.w=sWidth; r.h=sHeight;

  if(useRootWindow) {
#ifdef WITH_ROOTWINDOW
    if ((display = XOpenDisplay(0)) == 0)
      return -1;
    x11screen = DefaultScreen(display);

    if(!window) window=RootWindow(display,x11screen);

    XWindowAttributes wa;
    XGetWindowAttributes(display, window, &wa);
    Visual *visual = wa.visual;
    XVisualInfo templ;
    templ.screen = x11screen;
    templ.visualid = XVisualIDFromVisual(visual);
    int out_count;
    XVisualInfo *vinfo = XGetVisualInfo(display,
					VisualScreenMask | VisualIDMask, &templ, &out_count);
    if (!vinfo)
      return -1;  
    GLXContext context = glXCreateContext(display, vinfo, 0, GL_TRUE);
    XFree(vinfo); 
    if (!context)
      return -1;  
    if (!glXMakeCurrent(display, window, context))
      return -1;    
    XMapRaised(display, window);    
    XWindowAttributes attrib;
    XGetWindowAttributes(display, window, &attrib);
    screenWidth = attrib.width;
    screenHeight = attrib.height;
    //printf("screenWidth=%d\nscreenHeight=%d\n",screenWidth,screenHeight);
    // Initialze SDL
    if((SDL_Init(0)==-1)) {
      printf("Could not initialize SDL: %s.\n", SDL_GetError());
      exit(-1);
    }  
    atexit(SDL_Quit);
    printf("GLX initialized\n");
#else
    printf("Compiled with no root window support\n");
#endif
  } else {
    /* Initialize defaults, Video and Audio */
    if((SDL_Init(SDL_INIT_VIDEO)==-1)) { 
      printf("Could not initialize SDL: %s.\n", SDL_GetError());
      exit(-1);
    }
    
    /* Clean up on exit */
    atexit(SDL_Quit);
    
    //  screen = SDL_SetVideoMode(sWidth, sHeight, 24, SDL_SWSURFACE | SDL_DOUBLEBUF | (isWindowed?0:SDL_FULLSCREEN));
    screen = SDL_SetVideoMode(sWidth, sHeight, 16, SDL_SWSURFACE | SDL_OPENGL | (isWindowed?0:SDL_FULLSCREEN));
    if(!screen) { printf("Failed to initialize SDL\n"); exit(0); }
    screenWidth = screen->w; screenHeight = screen->h;

    SDL_ShowCursor(SDL_DISABLE);
    /*printf("SDL initialized: %dx%d@%dbpp\n",screen->w,screen->h,screen->format->BitsPerPixel);*/
  }


  /* Read parameters from stdin */
  printf("Loading parameters from %s\n",settingsFile);
  yyin=fopen(settingsFile,"r");
  if(!yyin) { printf("Failed to open %s\n",settingsFile); exit(0); }
  yyparse();
  simulationRadius = (m_width/2.0/m_res);
  for(i=0;i<MAX_POINTS;i++) {
    points[i].index = i;
    points[i].heat = 0;
    points[i].burn_left = burn_steps;
    points[i].is_burning=0;    
    points[i].angular=0.0;    
    for(j=0;j<N_LINKS;j++)
      points[i].links[j]=-1;
  }

  /* try to make behaviour independent of number of points simulated */
  gravity /= npoints / 1000.0;
  /*radiation /= npoints / 1000.0;*/
  /*friction_heat /= npoints / 1000.0;*/
  fusion_radiation /= npoints / 1000.0;
  fusion_chance /= npoints / 1000.0;

  is_running=1;

  /* Setup all slave threads */
  pthread_attr_t attr;
  pthread_attr_init(&attr);
  for(i=0;i<numSlaves;i++)
    pthread_create(&slaveThreads[i],&attr,slave_thread_fn,(void*)&i);

  /* Keep track of frames */
  frames=0;
  gettimeofday(&tv,NULL);
  t0 = (double) tv.tv_sec + 1e-6 * tv.tv_usec;
  double lastFrame=t0;

  /* Fast forward to end of playback if requested */
  if(replay_from_disk && skip_to_end) {
    fseek(file,-sizeof(DiskPoint)*npoints*2,SEEK_END);
  }

  workQueue=NULL;
  freeOrders=NULL;

  while(is_running) {
    frames++;
    gettimeofday(&tv,NULL);
    t1 = (double) tv.tv_sec + 1e-6 * tv.tv_usec;
    while(t1 - lastFrame < 1.0/30.0) { 
      usleep(1000);
      gettimeofday(&tv,NULL);
      t1=(double) tv.tv_sec + 1e-6 * tv.tv_usec;
    }
    lastFrame=t1;

    if((frames % 50) == 0) {
      if(printFPS) 
	printf("time per frame: %2.2f sec (%2.1f fps)\n",(t1-t0)/50.0,50.0/(t1-t0));
      t0 = t1;
    }

    offset_mx = 1.0 * frandom();
    offset_my = 1.0 * frandom();

    /* Wait until all slave threads are finished */
    while(workQueue != NULL) usleep(1000);

    /* rebuild map of points */
    for(x=0;x<m_width;x++)
      for(y=0;y<m_height;y++) {
	map[x][y].points = NULL;
	map[x][y].gravity = 0;
	map[x][y].heat = 0.0;
      }

	for(i=0;i<npoints;i++) {
	  x=m_width/2 + (m_res * points[i].pos[0] + offset_mx);
	  y=m_height/2 + (m_res * points[i].pos[1] + offset_my);
	  if(x >= 0 && x < m_width && y >= 0 && y < m_height) {
		points[i].next = map[x][y].points;
		map[x][y].points = &points[i];
		map[x][y].gravity++;
		map[x][y].heat += (points[i].heat+heat_off+fusion_radiation*points[i].is_burning*points[i].burn_left);
	  } /* else.. ok to ignore faraway points, they will be pushed in manually */
	}

	for(x=0;x<m_width/4;x++)
	  for(y=0;y<m_height/4;y++) {
		map2[x][y]=0;
		for(x2=x*4;x2<x*4+4;x2++)
		  for(y2=y*4;y2<y*4+4;y2++)
			map2[x][y] += map[x2][y2].gravity;
	  }
	for(x=0;x<m_width/16;x++)
	  for(y=0;y<m_height/16;y++) {
		map3[x][y]=0;
		for(x2=x*4;x2<x*4+4;x2++)
		  for(y2=y*4;y2<y*4+4;y2++)
			map3[x][y] += map2[x2][y2];
	  }

	/* stop the burning from previous turns */
	for(i=0;i<npoints;i++)
	  points[i].is_burning = 0;

	if(replay_from_disk) {
	  int s = fread(&diskpoints[0],sizeof(DiskPoint)*npoints,1,file);
	  if(s != 1) {
		replay_from_disk=0;
		record_to_disk=1;
		fclose(file);
		file=fopen(filename,"a");
	  } else {
		for(i=0;i<npoints;i++) {
		  points[i].dir[0] = diskpoints[i].pos[0] - points[i].pos[0];
		  points[i].dir[1] = diskpoints[i].pos[1] - points[i].pos[1];
		  points[i].pos[0] = diskpoints[i].pos[0];
		  points[i].pos[1] = diskpoints[i].pos[1];
		  points[i].heat = diskpoints[i].heat;
		  points[i].burn_left = diskpoints[i].burn_left;
		  points[i].is_burning = diskpoints[i].is_burning;
		}
	  }
	} else {
	  for(mx3=0;mx3<m_width/16;mx3++)
	    for(my3=0;my3<m_height/16;my3++)
	      if(map3[mx3][my3])
		for(mx2=mx3*4;mx2<mx3*4+4;mx2++)
		  for(my2=my3*4;my2<my3*4+4;my2++)
		    if(map2[mx2][my2])
		      for(x=mx2*4;x<mx2*4+4;x++)
			for(y=my2*4;y<my2*4+4;y++) 
			  if(map[x][y].points)
			    for(mx3_2=0;mx3_2<m_width/16;mx3_2++)
			      for(my3_2=0;my3_2<m_height/16;my3_2++)
				if(map3[mx3_2][my3_2])
				  for(mx2_2=mx3_2*4;mx2_2<mx3_2*4+4;mx2_2++)
				    for(my2_2=my3_2*4;my2_2<my3_2*4+4;my2_2++)
				      if(map2[mx2_2][my2_2])
					for(x2=mx2_2*4;x2<mx2_2*4+4;x2++)
					  for(y2=my2_2*4;y2<my2_2*4+4;y2++) 
					    if(map[x2][y2].points) {
					      /*d=(x2-x)*(x2-x)+(y2-y)*(y2-y);*/
					      if(abs(x2-x) + abs(y2-y) <= m_complicated) {
						/* do complicated interactions */
						queue_interaction(x,y,x2,y2);
						/*
						for(p1=map[x][y].points;p1;p1=p1->next)
						  for(p2=map[x2][y2].points;p2;p2=p2->next)
						    if(p2 > p1)
						    queue_interaction(p1,p2);*/
					      } else {
						double sqrt_d;
						
						/* do only simple gravity and heat radiation */
						dx = ((double)(x2 - x)) / m_res;
						dy = ((double) (y2 - y)) / m_res;
						d = dx*dx+dy*dy;
						sqrt_d = sqrt(d);
						dx = map[x2][y2].gravity * gravity * dx / (dist_off_g + d * sqrt_d);
						dy = map[x2][y2].gravity * gravity * dy / (dist_off_g + d * sqrt_d);
						d = transfer * radiation * 5.0 / sqrt_d * map[x2][y2].heat;
						
						for(p1=map[x][y].points;p1;p1=p1->next) {
						  p1->dir[0] += dx;
						  p1->dir[1] += dy;
						  p1->heat += d;
						}
					      }
					    }
	  /* loose heat */
	  for(i=0;i<npoints;i++)
	    points[i].heat *= 1.0 - radiation;
	  
	  /* friction on faraway objects */
	  for(i=0;i<npoints;i++) {
		dx=points[i].pos[0];
		dy=points[i].pos[1];
		d=dx*dx+dy*dy;
		if(d > simulationRadius*simulationRadius) {
		  double overlap=(sqrt(d)-simulationRadius) / simulationRadius;
		  points[i].dir[0] = points[i].dir[0]*(1.0 - 0.1*overlap) - 1e-3 * overlap * dx;
		  points[i].dir[1] = points[i].dir[1]*(1.0 - 0.1*overlap) - 1e-3 * overlap * dy;
		}
	  }

	  /* Effect of links, maintain links */
	  for(i=0;i<npoints;i++) 
	    for(j=0;j<N_LINKS;j++) {
	      p1 = &points[i];
	      if(p1->links[j] >= 0 && p1->links[j] < npoints) {
		p2 = &points[p1->links[j]];
		dx = p2->pos[0] - p1->pos[0];
		dy = p2->pos[1] - p1->pos[1];
		d = sqrt(dx*dx+dy*dy);
		if(d > max_link_distance) 
		  p1->links[j] = -1;                                      /* Break link */
		else {
		  double v;

		  /* Link propagates temperature */
		  v = (1.0 - link_temp_prop) * p1->heat + p2->heat * link_temp_prop;
		  p2->heat = (1.0 - link_temp_prop) * p2->heat + p1->heat * link_temp_prop;
		  p1->heat=v;

		  /* Links causes friction */
		  v = p1->dir[0] * (1.0 - link_fric) + p2->dir[0]*link_fric;
		  p2->dir[0] = p2->dir[0] * (1.0 - link_fric) + p1->dir[0]*link_fric;
		  p1->dir[0]=v;
		  v = p1->dir[1] * (1.0 - link_fric) + p2->dir[1]*link_fric;
		  p2->dir[1] = p2->dir[1] * (1.0 - link_fric) + p1->dir[1]*link_fric;
		  p1->dir[1]=v;
		  double dx2 = p1->dir[0] - p2->dir[0];
		  double dy2 = p2->dir[0] - p1->dir[0];
		  /* TODO - update to match other friction expression */
		  v=dx2*dx2+dy2*dy2 * friction_heat * 2.0 / (dist_off_f + d*d);
		  p1->heat += v;
		  p2->heat += v;

		  /* Calculate anglar momentum */
#ifdef OLD
		  /* 0. Shared velocity frame */
		  double frame_vx = (p1->dir[0] + p2->dir[0])/2;
		  double frame_vy = (p1->dir[1] + p2->dir[1])/2;
		  /* 1. Vector ortogonal to link */
		  double ortox = -dy, ortoy = dx, ortolen_sq=ortox*ortox+ortoy*ortoy;
		  /* 2. Current momentum of link (radians/sec)*/
		  double omega = ((p2->dir[0]-frame_vx) * ortox + (p2->dir[1]-frame_vy) * ortoy)/ortolen_sq;
		  /* 3. Apply force depending on angular momentum
		     differences and add part of angular momementums to points */
		  double ang_diff=angular_epsilon * (omega - p2->angular);
		  p2->angular += ang_diff;
		  p1->dir[0] += ortox * ang_diff;
		  p2->dir[0] -= ortox * ang_diff;
		  p1->dir[1] += ortoy * ang_diff;
		  p2->dir[1] -= ortoy * ang_diff;
		  ang_diff=angular_epsilon * (omega - p1->angular);
		  p1->angular += ang_diff;
		  p2->dir[0] += ortox * ang_diff;
		  p1->dir[0] -= ortox * ang_diff;
		  p2->dir[1] += ortoy * ang_diff;
		  p1->dir[1] -= ortoy * ang_diff;

		  /* Link acts as a spring, only if distance sufficiently large */
		  if(1 || d > link_spring_distance) {  
		    v = (d-link_spring_distance) * link_spring_constant;
		    p1->dir[0] += dx*v;
		    p1->dir[1] += dy*v;
		    p2->dir[0] -= dx*v;
		    p2->dir[1] -= dy*v;
		  }
#endif



		}		  
	      }
	    }

	  /* Move points */
	  for(i=0;i<npoints;i++) {
	    points[i].pos[0] += points[i].dir[0];
	    points[i].pos[1] += points[i].dir[1];
	  }
	
	  /* Center screen */
	  dx=dy=0.;
	  for(i=0;i<npoints;i++) { dx += points[i].pos[0]; dy += points[i].pos[1]; }
	  dx/=npoints*10.0; dy/=npoints*10.0;
	  for(i=0;i<npoints;i++) { points[i].pos[0] -= dx; points[i].pos[1] -= dy; }
	  /* and velocities */
	  dx=dy=0.;
	  for(i=0;i<npoints;i++) { dx += points[i].dir[0]; dy += points[i].dir[1]; }
	  dx/=npoints; dy/=npoints;
	  for(i=0;i<npoints;i++) { points[i].dir[0] -= dx; points[i].dir[1] -= dy; }
	}


	/* record simulation to disk */
	if(record_to_disk) {
	  for(i=0;i<npoints;i++) {
		diskpoints[i].pos[0] = points[i].pos[0];
		diskpoints[i].pos[1] = points[i].pos[1];
		diskpoints[i].heat = points[i].heat;
		diskpoints[i].burn_left = points[i].burn_left;
		diskpoints[i].is_burning = points[i].is_burning;		
	  }
	  fwrite(&diskpoints[0],sizeof(DiskPoint),npoints,file);
	}

	/* Draw points */  
	if(1 || (frames % 2 == 0)) {
	  
	  glClearColor(0.2,0.2,0.2,1.);
	  /*glClearColor(1.0,1.0,1.0,1.);*/
	  glClear(GL_COLOR_BUFFER_BIT);
	  glDisable(GL_DEPTH_TEST);
	  glDisable(GL_CULL_FACE);
	  glDisable(GL_LIGHTING);
	  glEnable(GL_BLEND);
	  //glBlendFunc(GL_ONE, GL_SRC_ALPHA);

	  glMatrixMode(GL_PROJECTION);
	  glLoadIdentity();
	  glMatrixMode(GL_MODELVIEW);
	  glLoadIdentity();
	  glViewport(0, 0, screenWidth, screenHeight); //screen->w, screen->h);
	  glOrtho(0.0, (GLdouble)screenWidth, (GLdouble)screenHeight, 0.0, 0.0, 1.0);

	  glEnable(GL_POINT_SMOOTH);
	  glBlendFunc(GL_SRC_ALPHA,GL_ONE); //GL_ONE_MINUS_SRC_ALPHA);
	  glTranslatef(-cam[0]*sWidth,-cam[1]*sWidth,0.0);

	  GLfloat rgb[4];
	  
	  /* Draw links */
	  if(drawLinks) {
	    glBegin(GL_LINES); 
	    for(i=0;i<npoints;i++)  
	      for(j=0;j<N_LINKS;j++)
		if(points[i].links[j] > -1) {
		  p1 = &points[i];
		  p2 = &points[p1->links[j]];
		  int x1=p1->pos[0] * 250.0 * zoom + sWidth/2;
		  int y1=p1->pos[1] * 250.0 * zoom + sHeight/2;
		  int x2=p2->pos[0] * 250.0 * zoom + sWidth/2;
		  int y2=p2->pos[1] * 250.0 * zoom + sHeight/2;
		
		  int n,c,c2;
		  d = pow(p1->heat,0.2) * 0.2 * (double) NUM_COLORS;
		  c = (int) d; d -= c; c2 = c+1;
		  if(c >= NUM_COLORS-1) { c = NUM_COLORS-1; d=0.0; c2 = NUM_COLORS-1; }
		      
		  glColor4f(colors[c].rgb[0][0] * (1.0-d) + d * colors[c2].rgb[0][0], 
			    colors[c].rgb[0][1] * (1.0-d) + d * colors[c2].rgb[0][1],
			    colors[c].rgb[0][2] * (1.0-d) + d * colors[c2].rgb[0][2],
			    0.2);
		  glVertex2i(x1,y1);
		
		  d = pow(p2->heat,0.2) * 0.2 * (double) NUM_COLORS;
		  c = (int) d; d -= c; c2 = c+1;
		  if(c >= NUM_COLORS-1) { c = NUM_COLORS-1; d=0.0; c2 = NUM_COLORS-1; }
		
		  glColor4f(colors[c].rgb[0][0] * (1.0-d) + d * colors[c2].rgb[0][0], 
			    colors[c].rgb[0][1] * (1.0-d) + d * colors[c2].rgb[0][1],
			    colors[c].rgb[0][2] * (1.0-d) + d * colors[c2].rgb[0][2],
			    0.2);
		  glVertex2i(x2,y2);
		}
	    glEnd();
	  }
	  
	  int steps;
	  for(steps=1;steps<=5;steps++) {
	    glPointSize((steps*1.0+1.0)*zoom);
	    glBegin(GL_POINTS);

		/* Draw all standard points, bucket sorted for coolest first */
		for(i=0;i<npoints;i++) {
		  x=points[i].pos[0] * 250.0 * zoom + sWidth/2;
		  y=points[i].pos[1] * 250.0 * zoom + sHeight/2;
		  if(x >= 0 && x < sWidth && y >= 0 && y < sHeight) {
			int n,c,c2;
			d = pow(points[i].heat,0.2) * 0.2 * (double) NUM_COLORS;
			c = (int) d; d -= c; c2 = c+1;
			if(c >= NUM_COLORS-1) { c = NUM_COLORS-1; d=0.0; c2 = NUM_COLORS-1; }
			n=colors[c2].n;
			if(n != steps) continue;
			
			rgb[0] = colors[c].rgb[0][0] * (1.0-d) + d * colors[c2].rgb[0][0];
			rgb[1] = colors[c].rgb[0][1] * (1.0-d) + d * colors[c2].rgb[0][1];
			rgb[2] = colors[c].rgb[0][2] * (1.0-d) + d * colors[c2].rgb[0][2];
			rgb[3] = colors[c].rgb[0][3] * (1.0-d) + d * colors[c2].rgb[0][3];
			glColor4fv(rgb);
			glVertex2i(x,y);
		  }
		}
		glEnd();
		
		/* Draw burning points */
		glPointSize(1.0*zoom);
		glBegin(GL_POINTS);
		for(i=0;i<npoints;i++) {
		  x=points[i].pos[0] * 250.0 * zoom + sWidth/2;
		  y=points[i].pos[1] * 250.0 * zoom + sHeight/2;
		  if(x >= 0 && x < sWidth && y >= 0 && y < sHeight && points[i].is_burning) {
		    int bc=burn_steps - points[i].burn_left; if(bc>=MAX_BURN_STEPS) bc=MAX_BURN_STEPS-1;		    
		    glColor4fv(burn_colors[bc]);
		    glVertex2i(x,y);			
		    if(x >= 20 && x < sWidth-20 && y >= 20 && y < sHeight - 20) {
		      for(j=0;j<10+points[i].burn_left;j++) {
			double v=(10+points[i].burn_left-j)/20.0;
			glColor4f(burn_colors[bc][0]*v,burn_colors[bc][1]*v,burn_colors[bc][2]*v,burn_colors[bc][3]*v);
			glVertex2i(x+j,y);
			glVertex2i(x-j,y);
			glVertex2i(x,y+j);
			glVertex2i(x,y-j);
		      }			  
		    }
		  }
		}
		glEnd();

		
	  }

	  if(useRootWindow) 
	    glXSwapBuffers(display, window);
	  else
	    SDL_GL_SwapBuffers();
	}
	
	if(!useRootWindow) {
	  while(SDL_PollEvent(&event)) 
	    switch(event.type) {
	    case SDL_QUIT: is_running=0; break;
	    case SDL_KEYDOWN: 
	      if(event.key.keysym.sym == 'a') zoom *= 0.8;
	      if(event.key.keysym.sym == 's') zoom /= 0.8;
	      if(event.key.keysym.sym == 'c') createComet();
	      if(event.key.keysym.sym == 'p') createPlanet();
	      if(event.key.keysym.sym == SDLK_ESCAPE) is_running=0; 
	      
	      if(event.key.keysym.sym == SDLK_UP) cam[1] -= zoom * 0.05;
	      else if(event.key.keysym.sym == SDLK_DOWN) cam[1] += zoom * 0.05;
	      if(event.key.keysym.sym == SDLK_LEFT) cam[0] -= zoom * 0.05;
	      else if(event.key.keysym.sym == SDLK_RIGHT) cam[0] += zoom * 0.05;
	      
	      break;
	    }
	}
  }

  if(record_to_disk || replay_from_disk)
	fclose(file);

  /* Shutdown all subsystems */
  SDL_Quit();
  exit(0);
}

void interact(Point *p1,Point *p2) {
  double dx,dy,d;
  double dr,sqrt_d;
  double rep;
  
  /* Gravity */
  dx = p2->pos[0] - p1->pos[0];
  dy = p2->pos[1] - p1->pos[1];
  d = dx*dx+dy*dy;
  sqrt_d = sqrt(d);

  p1->dir[0] += gravity * dx / (dist_off_g + d * sqrt_d);
  p1->dir[1] += gravity * dy / (dist_off_g + d * sqrt_d);
  p2->dir[0] -= gravity * dx / (dist_off_g + d * sqrt_d);
  p2->dir[1] -= gravity * dy / (dist_off_g + d * sqrt_d);

  /* Create links */
  if(sqrt_d < link_creation_distance && frandom() < link_creation_chance) { 
    p1->links[rand()%N_LINKS] = p2->index;
  }


  /* fusion */
  if(sqrt_d < fusion_dist*(2+p1->burn_left) && p1->burn_left > 0 && p1->burn_left == p2->burn_left) {
    //dr = p1->burn_left * p1->heat - 5.0; 
    dr = (p1->heat - 0.0 - 1.0 * p1->burn_left)/(10.0+2.0*p1->burn_left); if(dr > 1.0) dr = 1.0;
    if(frandom() < fusion_chance * dr / (fusion_dist + sqrt_d)) {
      printf("%1d",p1->burn_left);
      p1->burn_left--;
      p1->is_burning=1;
      p1->heat += fusion_heat * (burn_steps-p1->burn_left) / (0.5 + frandom());
    }
    //dr = p2->burn_left * p2->heat - 5.0; 
    dr = (p2->heat - 0.0 - 1.0 * p2->burn_left)/(10.0+2.0*p1->burn_left); if(dr > 1.0) dr = 1.0;
    if(frandom() < fusion_chance * dr / (fusion_dist + sqrt_d)) {
      printf("%1d",p2->burn_left);
      p2->burn_left--;
      p2->is_burning=1;
      p2->heat += fusion_heat * (burn_steps-p1->burn_left) / (0.5 + frandom());
    }
  }

  /* heat transfer */
  dr = transfer * radiation / (d+1e-4) * (p2->heat+heat_off+fusion_radiation*p2->is_burning*p2->burn_left);
  p2->heat += transfer * radiation / (d+1e-4) * (p1->heat+heat_off+fusion_radiation*p1->is_burning*p1->burn_left);
  p1->heat += dr;

  /* volume */
  //rep = repell * (1.0 + 2.0 * pow(p1->heat + p2->heat,0.35));
  //double volume = 1.0 + 2.0 * pow(p1->heat + p2->heat,0.35);

  /* Compute pressure between particles, using the PV=T rule */
  double volume=d; /* Remember, d is here distance squared, ignoring PI and other constants */
  double pressure = (4.0 + 1.0 * pow(p1->heat,0.26) + pow(p2->heat,0.26)) / (dist_off_r + volume);

  /* Apply repelation forces if particles are sufficiently close */
  if(sqrt_d < 4e-4) {
    p1->dir[0] += repell * pressure * dx / (dist_off_r + sqrt_d);
    p1->dir[1] += repell * pressure * dy / (dist_off_r + sqrt_d);
    p2->dir[0] -= repell * pressure * dx / (dist_off_r + sqrt_d);
    p2->dir[1] -= repell * pressure * dy / (dist_off_r + sqrt_d);
  }

  /*
  rep = repell * volume;
  p1->dir[0] += scale * rep * dx / (dist_off_r + d * sqrt_d);
  p1->dir[1] += scale * rep * dy / (dist_off_r + d * sqrt_d);
  p2->dir[0] -= scale * rep * dx / (dist_off_r + d * sqrt_d);
  p2->dir[1] -= scale * rep * dy / (dist_off_r + d * sqrt_d);
  */

  /* Apply friction if particles are sufficiently close - this
     deletes energy but keeps momentum in the system */
  if(sqrt_d < 4e-4) {
    /* Friction */
    dx = p2->dir[0] - p1->dir[0];
    dy = p2->dir[1] - p1->dir[1];
    sqrt_d = sqrt(dx*dx+dy*dy);
    dx = dx * friction / (dist_off_f + sqrt_d);// Was * pressure
    dy = dy * friction / (dist_off_f + sqrt_d); 
    p1->dir[0] += dx;
    p1->dir[1] += dy;
    p2->dir[0] -= dx;
    p2->dir[1] -= dy;
    p1->heat += (dx*dx+dy*dy) * friction_heat;
    p2->heat += (dx*dx+dy*dy) * friction_heat;
  }

#ifdef foo  
  if(sqrt_d < 1e-3*volume) {
	/* Friction */
	dx = p2->dir[0] - p1->dir[0];
	dy = p2->dir[1] - p1->dir[1];
	sqrt_d = sqrt(dx*dx+dy*dy);
	p1->dir[0] += scale * dx * friction / (dist_off_f + d * d);
	p1->dir[1] += scale * dy * friction / (dist_off_f + d * d);
	p2->dir[0] -= scale * dx * friction / (dist_off_f + d * d);
	p2->dir[1] -= scale * dy * friction / (dist_off_f + d * d);
	p1->heat += scale * (dx*dx+dy*dy) * friction_heat / (dist_off_f + d * d);
	p2->heat += scale * (dx*dx+dy*dy) * friction_heat / (dist_off_f + d * d);
  }		
#endif
}

int log2i(int x) {
  int i;
  for(i=0;i<31;i++)
	if(x<=(1<<i)) return i;
}

void sphere(int from,int to,double cx,double cy,double rad) {
  int i;
  double d,a;

  for(i=from;i<to;i++) {	
	d=frandom()*rad; a=frandom() * M_PI * 2.0;
	points[i].pos[0] = cx + d*sin(a);
	points[i].pos[1] = cy + d*cos(a);
  }
}
void cube(int from,int to,double cx,double cy,double rad) {
  int i;

  for(i=from;i<to;i++) {
	points[i].pos[0] = (frandom()*2.0-1.0)*rad+cx;
	points[i].pos[1] = (frandom()*2.0-1.0)*rad+cy;
  }
}
void spin(int from,int to,double cx,double cy,double rad,double speed) {
  int i;
  double dx,dy,d;

  for(i=from;i<to;i++) {
	dx = points[i].pos[0] - cx;
	dy = points[i].pos[1] - cy;
	d=dx*dx+dy*dy;
	if(sqrt(d) < rad) {
	  points[i].dir[0] += speed * dy; /* / (sqrt(d)+0.001*rad);*/
	  points[i].dir[1] -= speed * dx; /* / (sqrt(d)+0.001*rad);*/
	}
  }
}
void linear(int from,int to,double dx,double dy) {
  int i;
  for(i=from;i<to;i++) {
	points[i].dir[0] += dx;
	points[i].dir[1] += dy;
  }
}
void drift(int from,int to,double speed) {
  int i;
  for(i=from;i<to;i++) {
	points[i].dir[0] += speed * (frandom()*2.0-1.0);
	points[i].dir[1] += speed * (frandom()*2.0-1.0);
  }
}
void zero(int from,int to) {
  int i;
  for(i=from;i<to;i++) {
	points[i].dir[0] = 0.0;
	points[i].dir[1] = 0.0;
  }
}
/* returns a random number between 0.0 and 1.0 */
double frandom() { return (rand()%(1<<24))/((double)(1<<24)); }

/* uses some existing particles (2-10%) to create a new "comet" at the edge of the 
   simulation area moving inward */
void createComet( ){
  int i,n,cnt;
  n=(npoints*((rand()%9)+2))/100 + 1;
  double a=frandom()*2.0*M_PI;
  double r=simulationRadius*(0.8+0.2*frandom());
  double cx=r*sin(a), cy=r*cos(a);  
  for(cnt=0;cnt<n;cnt++) {
    i = rand() % npoints;
    double a2=frandom()*2.0*M_PI;
    double r2=(frandom()+0.001)*cometRadius;
    points[i].pos[0] = cx + sin(a2)*r2;
    points[i].pos[1] = cy + cos(a2)*r2;
    points[i].dir[0] = -cx*1e-2;
    points[i].dir[1] = -cy*1e-2;
    points[i].heat = 0.0;
  }
}

/* uses some existing particles (5-10%) to create a new "planet" */
void createPlanet( ){
  int i,n,cnt;
  n=(npoints*((rand()%6)+5))/100 + 1;
  double a=frandom()*2.0*M_PI;
  double r=simulationRadius*(0.8+0.2*frandom());
  double cx=r*sin(a), cy=r*cos(a);  
  double dx=(frandom()-0.5)*1e-1;
  double dy=(frandom()-0.5)*1e-1;
  for(cnt=0;cnt<n;cnt++) {
    i = rand() % npoints;
    double a2=frandom()*2.0*M_PI;
    double r2=(frandom()+0.001)*cometRadius;
    points[i].pos[0] = cx + sin(a2)*r2;
    points[i].pos[1] = cy + cos(a2)*r2;
    points[i].dir[0] = -cx*1e-2+dx;
    points[i].dir[1] = -cy*1e-2+dy;
    points[i].heat = 0.0;
  }
}

void queue_interaction(int x1,int y1,int x2,int y2) {
  while(!freeOrders && maxOrders > 10000) usleep(10);

  pthread_mutex_lock(&orderMutex);
  if(!freeOrders) {
    freeOrders = (WorkOrder*) malloc(sizeof(WorkOrder));
    freeOrders->next=NULL;
    maxOrders++;
    /*printf("Max Orders: %d\n",maxOrders);*/
  }
  WorkOrder *order = freeOrders;
  freeOrders=freeOrders->next;
  order->next = workQueue;
  order->x1 = x1;
  order->y1 = y1;
  order->x2 = x2;
  order->y2 = y2;
  workQueue = order;
  pthread_mutex_unlock(&orderMutex); 
}

void *slave_thread_fn(void *arg){ 
  int id=*((int*) arg);
  while(is_running) {
    WorkOrder *order;
    Point *p1, *p2;
    pthread_mutex_lock(&orderMutex);
    order=workQueue;
    if(workQueue) workQueue = workQueue->next;
    pthread_mutex_unlock(&orderMutex); 
    if(!order) usleep(1);
    else {
      
      /* Do all interactions between these two cells */
      for(p1=map[order->x1][order->y1].points;p1;p1=p1->next)
	for(p2=map[order->x2][order->y2].points;p2;p2=p2->next)
	  if(p2>p1) interact(p1,p2);

      pthread_mutex_lock(&orderMutex);
      order->next = freeOrders;
      freeOrders = order;
      pthread_mutex_unlock(&orderMutex);       
    }
  }
  return NULL;
}
